@@ -27,11 +27,13 @@
 #endif
 
 #include "fs-rtp-codec-negotiation.h"
-#include "fs-rtp-specific-nego.h"
-#include "fs-rtp-conference.h"
 
 #include <string.h>
 
+#include "fs-rtp-codec-specific.h"
+#include "fs-rtp-conference.h"
+
+
 #define GST_CAT_DEFAULT fsrtpconference_nego
 
 #define SEND_PROFILE_ARG "farsight-send-profile"
@@ -46,7 +48,7 @@ static CodecAssociation *
 codec_association_copy (CodecAssociation *ca);
 
 static CodecAssociation *
-lookup_codec_association_custom_intern (GList *codec_associations,
+lookup_codec_association_custom_internal (GList *codec_associations,
     gboolean want_disabled, CAFindFunc func, gpointer user_data);
 
 static gboolean
@@ -126,16 +128,13 @@ find_matching_pad (gconstpointer a, gconstpointer b)
   GstPad *pad = GST_PAD (a);
   GstCaps *caps = GST_CAPS (b);
   GstCaps *padcaps = NULL;
-  GstCaps *intersect = NULL;
   gint ret = 1;
 
-  padcaps = gst_pad_get_caps (pad);
-  intersect = gst_caps_intersect (caps, padcaps);
+  padcaps = gst_pad_get_caps_reffed (pad);
 
-  if (intersect && !gst_caps_is_empty (intersect))
+  if (gst_caps_can_intersect (caps, padcaps))
     ret = 0;
 
-  gst_caps_unref (intersect);
   gst_caps_unref (padcaps);
   gst_object_unref (pad);
 
@@ -216,6 +215,20 @@ validate_codec_profile (FsCodec *codec,const gchar *bin_description,
   return TRUE;
 }
 
+static gboolean
+codec_sdp_compare (FsCodec *local_codec, FsCodec *remote_codec)
+{
+  FsCodec *nego_codec = sdp_negotiate_codec (
+      local_codec, FS_PARAM_TYPE_ALL & ~FS_PARAM_TYPE_CONFIG,
+      remote_codec, FS_PARAM_TYPE_ALL & ~FS_PARAM_TYPE_CONFIG);
+
+  if (!nego_codec)
+    return FALSE;
+
+  fs_codec_destroy (nego_codec);
+  return TRUE;
+}
+
 /**
  * validate_codecs_configuration:
  * @media_type: The #FsMediaType these codecs should be for
@@ -255,7 +268,6 @@ validate_codecs_configuration (FsMediaType media_type, GList *blueprints,
          blueprint_e = g_list_next (blueprint_e))
     {
       CodecBlueprint *blueprint = blueprint_e->data;
-      GList *codecparam_e = NULL;
 
       /* First, lets check the encoding name */
       if (g_ascii_strcasecmp (blueprint->codec->encoding_name,
@@ -269,34 +281,8 @@ validate_codecs_configuration (FsMediaType media_type, GList *blueprints,
       else if (!blueprint->codec->clock_rate && !codec->clock_rate)
         continue;
 
-      /* Now lets check that all params that are present in both
-       * match
-       */
-      for (codecparam_e = codec->optional_params;
-           codecparam_e;
-           codecparam_e = g_list_next (codecparam_e))
-      {
-        FsCodecParameter *codecparam = codecparam_e->data;
-        GList *bpparam_e = NULL;
-        for (bpparam_e = blueprint->codec->optional_params;
-             bpparam_e;
-             bpparam_e = g_list_next (bpparam_e))
-        {
-          FsCodecParameter *bpparam = bpparam_e->data;
-          if (!g_ascii_strcasecmp (codecparam->name, bpparam->name))
-          {
-            /* If the blueprint and the codec specify the value
-             * of a parameter, they should be the same
-             */
-            if (g_ascii_strcasecmp (codecparam->value, bpparam->value))
-              goto next_blueprint;
-            break;
-          }
-        }
-      }
-      break;
-    next_blueprint:
-      continue;
+      if (codec_sdp_compare (blueprint->codec, codec))
+        break;
     }
 
     /* If there are send and/or recv profiles, lets test them */
@@ -313,7 +299,6 @@ validate_codecs_configuration (FsMediaType media_type, GList *blueprints,
     {
       /* Accept codecs with no blueprints if they have a valid profile */
       if (fs_codec_get_optional_parameter (codec, RECV_PROFILE_ARG, NULL) &&
-          codec->id >= 0 && codec->id < 128 &&
           codec->encoding_name && codec->clock_rate)
         goto accept_codec;
 
@@ -348,6 +333,7 @@ _codec_association_destroy (CodecAssociation *ca)
     return;
 
   fs_codec_destroy (ca->codec);
+  fs_codec_destroy (ca->send_codec);
   g_free (ca->send_profile);
   g_free (ca->recv_profile);
   g_slice_free (CodecAssociation, ca);
@@ -373,17 +359,8 @@ _find_matching_blueprint (FsCodec *codec, GList *blueprints)
   for (item = g_list_first (blueprints); item; item = g_list_next (item))
   {
     CodecBlueprint *bp = item->data;
-    GstCaps *intersectedcaps = NULL;
-    gboolean ok = FALSE;
 
-    intersectedcaps = gst_caps_intersect (caps, bp->rtp_caps);
-
-    if (!gst_caps_is_empty (intersectedcaps))
-      ok = TRUE;
-
-    gst_caps_unref (intersectedcaps);
-
-    if (ok)
+    if (gst_caps_can_intersect (caps, bp->rtp_caps))
       break;
   }
 
@@ -422,7 +399,6 @@ _is_disabled (GList *codec_prefs, CodecBlueprint *bp)
   for (item = g_list_first (codec_prefs); item; item = g_list_next (item))
   {
     FsCodec *codec = item->data;
-    GstCaps *intersectedcaps = NULL;
     GstCaps *caps = NULL;
     gboolean ok = FALSE;
 
@@ -434,12 +410,9 @@ _is_disabled (GList *codec_prefs, CodecBlueprint *bp)
     if (!caps)
       continue;
 
-    intersectedcaps = gst_caps_intersect (caps, bp->rtp_caps);
-
-    if (!gst_caps_is_empty (intersectedcaps))
+    if (gst_caps_can_intersect (caps, bp->rtp_caps))
       ok = TRUE;
 
-    gst_caps_unref (intersectedcaps);
     gst_caps_unref (caps);
 
     if (ok)
@@ -458,14 +431,8 @@ static gboolean
 match_original_codec_and_codec_pref (CodecAssociation *ca, gpointer user_data)
 {
   FsCodec *codec_pref = user_data;
-  FsCodec *tmpcodec = NULL;
 
-  tmpcodec = sdp_is_compat (codec_pref, ca->codec, FALSE);
-
-  if (tmpcodec)
-    fs_codec_destroy (tmpcodec);
-
-  return (tmpcodec != NULL);
+  return codec_sdp_compare (codec_pref, ca->codec);
 }
 
 static void
@@ -540,10 +507,13 @@ create_local_codec_associations (
   GList *codec_pref_e = NULL;
   GList *lca_e = NULL;
   gboolean has_valid_codec = FALSE;
+  CodecAssociation *oldca = NULL;
 
   if (blueprints == NULL)
     return NULL;
 
+  GST_DEBUG ("Creating local codec associations");
+
   /* First, lets create the original table by looking at our preferred codecs */
   for (codec_pref_e = codec_prefs;
        codec_pref_e;
@@ -586,53 +556,56 @@ create_local_codec_associations (
     /* Now lets see if there is an existing codec that matches this preference
      */
 
+    if (codec_pref->id == FS_CODEC_ID_ANY)
     {
-      CodecAssociation *oldca = NULL;
+      oldca = lookup_codec_association_custom_internal (
+          current_codec_associations, TRUE,
+          match_original_codec_and_codec_pref, codec_pref);
+    }
+    else
+    {
+      oldca = lookup_codec_association_by_pt_list (current_codec_associations,
+          codec_pref->id, FALSE);
+      if (oldca && oldca->reserved)
+        oldca = NULL;
+    }
 
-      if (codec_pref->id == FS_CODEC_ID_ANY)
+    /* In this case, we have a matching codec association, lets keep the
+     * payload type from it
+     */
+    if (oldca)
+    {
+      FsCodec *codec = sdp_negotiate_codec (
+          oldca->codec, FS_PARAM_TYPE_BOTH | FS_PARAM_TYPE_CONFIG,
+          codec_pref, FS_PARAM_TYPE_ALL);
+      FsCodec *send_codec;
+
+      if (codec)
       {
-        oldca = lookup_codec_association_custom_intern (
-            current_codec_associations, TRUE,
-            match_original_codec_and_codec_pref, codec_pref);
+        fs_codec_destroy (codec);
+
+        send_codec = sdp_negotiate_codec (
+            oldca->send_codec, FS_PARAM_TYPE_SEND,
+            codec_pref, FS_PARAM_TYPE_SEND | FS_PARAM_TYPE_SEND_AVOID_NEGO);
+        if (send_codec)
+          fs_codec_destroy (send_codec);
+        else
+          oldca = NULL;
       }
       else
       {
-        oldca = lookup_codec_association_by_pt_list (current_codec_associations,
-            codec_pref->id, FALSE);
-        if (oldca && oldca->reserved)
-          oldca = NULL;
-      }
-
-      /* In this case, we have a matching codec association, lets keep it */
-      if (oldca)
-      {
-        FsCodec *codec = sdp_is_compat (codec_pref, oldca->codec, FALSE);
-        if (codec)
-        {
-          ca = g_slice_new (CodecAssociation);
-          memcpy (ca, oldca, sizeof (CodecAssociation));
-          codec_remove_parameter (codec, SEND_PROFILE_ARG);
-          codec_remove_parameter (codec, RECV_PROFILE_ARG);
-          codec->ABI.ABI.maxptime = codec_pref->ABI.ABI.maxptime;
-          codec->ABI.ABI.ptime = codec_pref->ABI.ABI.ptime;
-          ca->codec = codec;
-
-          ca->send_profile = dup_param_value (codec_pref, SEND_PROFILE_ARG);
-          ca->recv_profile = dup_param_value (codec_pref, RECV_PROFILE_ARG);
-
-          codec_associations = list_insert_local_ca (codec_associations, ca);
-          continue;
-        }
+        oldca = NULL;
       }
     }
 
     ca = g_slice_new0 (CodecAssociation);
     ca->blueprint = bp;
     ca->codec = fs_codec_copy (codec_pref);
-
     codec_remove_parameter (ca->codec, SEND_PROFILE_ARG);
     codec_remove_parameter (ca->codec, RECV_PROFILE_ARG);
-
+    ca->send_codec = codec_copy_filtered (ca->codec, FS_PARAM_TYPE_CONFIG);
+    if (oldca)
+      ca->send_codec->id = ca->codec->id = oldca->codec->id;
     ca->send_profile = dup_param_value (codec_pref, SEND_PROFILE_ARG);
     ca->recv_profile = dup_param_value (codec_pref, RECV_PROFILE_ARG);
 
@@ -642,7 +615,9 @@ create_local_codec_associations (
        * The blueprint has its own id, lets use it */
       if (ca->codec->id == FS_CODEC_ID_ANY &&
           (bp->codec->id >= 0 || bp->codec->id < 128))
-        ca->codec->id = bp->codec->id;
+      {
+        ca->send_codec->id = ca->codec->id = bp->codec->id;
+      }
 
       if (ca->codec->clock_rate == 0)
         ca->codec->clock_rate = bp->codec->clock_rate;
@@ -654,17 +629,9 @@ create_local_codec_associations (
            bp_param_e;
            bp_param_e = g_list_next (bp_param_e))
       {
-        GList *pref_param_e = NULL;
         FsCodecParameter *bp_param = bp_param_e->data;
-        for (pref_param_e = ca->codec->optional_params;
-             pref_param_e;
-             pref_param_e = g_list_next (pref_param_e))
-        {
-          FsCodecParameter *pref_param = pref_param_e->data;
-          if (!g_ascii_strcasecmp (bp_param->name, pref_param->name))
-            break;
-        }
-        if (!pref_param_e)
+
+        if (fs_codec_get_optional_parameter (ca->codec, bp_param->name, NULL))
           fs_codec_add_optional_parameter (ca->codec, bp_param->name,
               bp_param->value);
       }
@@ -693,7 +660,7 @@ create_local_codec_associations (
 
     if (lca->codec->id < 0)
     {
-      lca->codec->id = _find_first_empty_dynamic_entry (
+      lca->send_codec->id = lca->codec->id = _find_first_empty_dynamic_entry (
           current_codec_associations, codec_associations);
       if (lca->codec->id < 0)
       {
@@ -759,6 +726,7 @@ create_local_codec_associations (
         ca->blueprint = bp;
         ca->codec = fs_codec_copy (bp->codec);
         ca->codec->id = tmpca->codec->id;
+        ca->send_codec = codec_copy_filtered (ca->codec, FS_PARAM_TYPE_CONFIG);
 
         codec_associations = list_insert_local_ca (codec_associations, ca);
         next = TRUE;
@@ -782,6 +750,8 @@ create_local_codec_associations (
       }
     }
 
+    ca->send_codec = codec_copy_filtered (ca->codec, FS_PARAM_TYPE_CONFIG);
+
     codec_associations = list_insert_local_ca (codec_associations, ca);
   }
 
@@ -814,12 +784,43 @@ create_local_codec_associations (
   return NULL;
 }
 
+static void
+negotiate_stream_codec (CodecAssociation *old_ca, FsCodec *remote_codec,
+    gboolean multi_stream, FsCodec **nego_codec, FsCodec **nego_send_codec)
+{
+  if (multi_stream)
+    *nego_codec = sdp_negotiate_codec (old_ca->codec, FS_PARAM_TYPE_ALL,
+        remote_codec, FS_PARAM_TYPE_SEND | FS_PARAM_TYPE_SEND_AVOID_NEGO);
+  else
+    *nego_codec = sdp_negotiate_codec (old_ca->codec, FS_PARAM_TYPE_ALL,
+        remote_codec, FS_PARAM_TYPE_SEND);
+
+  if (*nego_codec)
+  {
+    if (multi_stream)
+      *nego_send_codec = sdp_negotiate_codec (
+          old_ca->send_codec,
+          FS_PARAM_TYPE_BOTH | FS_PARAM_TYPE_SEND_AVOID_NEGO,
+          remote_codec, FS_PARAM_TYPE_SEND | FS_PARAM_TYPE_SEND_AVOID_NEGO);
+    else
+      *nego_send_codec = sdp_negotiate_codec (
+          old_ca->send_codec, FS_PARAM_TYPE_BOTH,
+          remote_codec, FS_PARAM_TYPE_SEND | FS_PARAM_TYPE_SEND_AVOID_NEGO);
+
+    /* If send codec can't be negotiated, try another one */
+    if (!*nego_send_codec)
+    {
+      fs_codec_destroy (*nego_codec);
+      *nego_codec = NULL;
+    }
+  }
+}
+
 /**
  * negotiate_stream_codecs:
  * @remote_codecs: Remote codecs for the stream
  * @current_codec_assocations: The current list of #CodecAssociation
- * @use_local_ids: Whether to use local or remote PTs if they dont match (%TRUE
- *  for local, %FALSE for remote)
+ * @multi_stream: %TRUE if there is more than one stream.
  *
  * This function performs codec negotiation for a single stream. It does an
  * intersection of the current codecs and the remote codecs.
@@ -831,17 +832,21 @@ GList *
 negotiate_stream_codecs (
     const GList *remote_codecs,
     GList *current_codec_associations,
-    gboolean use_local_ids)
+    gboolean multi_stream)
 {
   GList *new_codec_associations = NULL;
   const GList *rcodec_e = NULL;
   GList *item = NULL;
 
+  GST_DEBUG ("Negotiating stream codecs (for %s)",
+      multi_stream ? "a single stream" : "multiple streams");
+
   for (rcodec_e = remote_codecs;
        rcodec_e;
        rcodec_e = g_list_next (rcodec_e)) {
     FsCodec *remote_codec = rcodec_e->data;
     FsCodec *nego_codec = NULL;
+    FsCodec *nego_send_codec = NULL;
     CodecAssociation *old_ca = NULL;
 
     gchar *tmp = fs_codec_to_string (remote_codec);
@@ -855,7 +860,8 @@ negotiate_stream_codecs (
 
     if (old_ca) {
       GST_DEBUG ("Have local codec in the same PT, lets try it first");
-      nego_codec = sdp_is_compat (old_ca->codec, remote_codec, FALSE);
+      negotiate_stream_codec (old_ca, remote_codec, multi_stream,
+          &nego_codec, &nego_send_codec);
     }
 
     if (!nego_codec) {
@@ -866,12 +872,16 @@ negotiate_stream_codecs (
       {
         old_ca = item->data;
 
-        nego_codec = sdp_is_compat (old_ca->codec, remote_codec, FALSE);
+        negotiate_stream_codec (old_ca, remote_codec, multi_stream,
+            &nego_codec, &nego_send_codec);
 
         if (nego_codec)
         {
-          if (use_local_ids)
-            nego_codec->id = old_ca->codec->id;
+          /* If we have multiple streams with codecs,
+           * then priorize the local IDs */
+          if (multi_stream)
+            nego_send_codec->id = nego_codec->id = old_ca->codec->id;
+
           break;
         }
       }
@@ -883,22 +893,11 @@ negotiate_stream_codecs (
 
       new_ca->need_config = old_ca->need_config;
       new_ca->codec = nego_codec;
+      new_ca->send_codec = nego_send_codec;
       new_ca->blueprint = old_ca->blueprint;
       new_ca->send_profile = g_strdup (old_ca->send_profile);
       new_ca->recv_profile = g_strdup (old_ca->recv_profile);
-      if (remote_codec->ABI.ABI.ptime && old_ca->ptime)
-        new_ca->ptime = MIN (remote_codec->ABI.ABI.ptime, old_ca->ptime);
-      else if (remote_codec->ABI.ABI.ptime)
-        new_ca->ptime = remote_codec->ABI.ABI.ptime;
-      else if (old_ca->ptime)
-        new_ca->ptime = old_ca->ptime;
-      if (remote_codec->ABI.ABI.maxptime && old_ca->maxptime)
-        new_ca->maxptime = MIN (remote_codec->ABI.ABI.maxptime,
-            old_ca->maxptime);
-      else if (remote_codec->ABI.ABI.maxptime)
-        new_ca->maxptime = remote_codec->ABI.ABI.maxptime;
-      else if (old_ca->maxptime)
-        new_ca->maxptime = old_ca->maxptime;
+
       tmp = fs_codec_to_string (nego_codec);
       GST_DEBUG ("Negotiated codec %s", tmp);
       g_free (tmp);
@@ -1007,7 +1006,8 @@ finish_codec_negotiation (
 
     if (old_ca)
     {
-      FsCodec *old_without_config = codec_copy_without_config (old_ca->codec);
+      FsCodec *old_without_config = codec_copy_filtered (old_ca->codec,
+          FS_PARAM_TYPE_CONFIG);
 
       if (fs_codec_are_equal (new_ca->codec, old_without_config))
       {
@@ -1111,6 +1111,7 @@ codec_association_copy (CodecAssociation *ca)
 
   memcpy (newca, ca, sizeof(CodecAssociation));
   newca->codec = fs_codec_copy (ca->codec);
+  newca->send_codec = fs_codec_copy (ca->send_codec);
   newca->send_profile = g_strdup (ca->send_profile);
   newca->recv_profile = g_strdup (ca->recv_profile);
 
@@ -1119,7 +1120,7 @@ codec_association_copy (CodecAssociation *ca)
 
 GList *
 codec_associations_to_codecs_internal (GList *codec_associations,
-    gboolean include_config, gboolean with_ptime)
+    gboolean include_config, gboolean send_codecs)
 {
   GList *codecs = NULL;
   GList *item = NULL;
@@ -1133,15 +1134,13 @@ codec_associations_to_codecs_internal (GList *codec_associations,
     {
       FsCodec *codec = NULL;
 
-      if (include_config)
+      if (send_codecs)
+        codec = fs_codec_copy (ca->send_codec);
+      else if (include_config)
         codec = fs_codec_copy (ca->codec);
       else
-        codec = codec_copy_without_config (ca->codec);
-      if (with_ptime)
-      {
-        codec->ABI.ABI.ptime = ca->ptime;
-        codec->ABI.ABI.maxptime = ca->maxptime;
-      }
+        codec = codec_copy_filtered (ca->codec, FS_PARAM_TYPE_CONFIG);
+
       codecs = g_list_append (codecs, codec);
     }
   }
@@ -1172,19 +1171,17 @@ codec_associations_to_codecs (GList *codec_associations,
 
 
 /**
- * codec_associations_to_codecs_with_ptime:
+ * codec_associations_to_send_codecs
  * @codec_associations: a #GList of #CodecAssociation
  *
  * Returns a #GList of the #FsCodec that are inside the list of associations
  * excluding those that are disabled or otherwise receive-only. It copies
- * the #FsCodec structures, but takes the ptime and maxptime from the
- * CodecAssociation. So it includes the ptime and maxptime that must be used
- * for sending.
+ * the #FsCodec structures. These codecs are to be used for sending
  *
  * Returns: a #GList of #FsCodec
  */
 GList *
-codec_associations_to_codecs_with_ptime (GList *codec_associations)
+codec_associations_to_send_codecs (GList *codec_associations)
 {
   return codec_associations_to_codecs_internal (codec_associations,
       FALSE, TRUE);
@@ -1194,7 +1191,8 @@ gboolean
 codec_association_is_valid_for_sending (CodecAssociation *ca,
     gboolean needs_codecbin)
 {
-  if (!ca->disable &&
+  if (ca->send_codec &&
+      !ca->disable &&
       !ca->reserved &&
       !ca->recv_only &&
       (!needs_codecbin ||
@@ -1208,7 +1206,7 @@ codec_association_is_valid_for_sending (CodecAssociation *ca,
 
 
 static CodecAssociation *
-lookup_codec_association_custom_intern (GList *codec_associations,
+lookup_codec_association_custom_internal (GList *codec_associations,
     gboolean want_disabled, CAFindFunc func, gpointer user_data)
 {
   GList *item;
@@ -1236,7 +1234,7 @@ lookup_codec_association_custom (GList *codec_associations,
     CAFindFunc func, gpointer user_data)
 {
 
-  return lookup_codec_association_custom_intern (codec_associations, FALSE,
+  return lookup_codec_association_custom_internal (codec_associations, FALSE,
       func, user_data);
 }
 
@@ -1309,29 +1307,16 @@ CodecAssociation *
 lookup_codec_association_by_codec_for_sending (GList *codec_associations,
     FsCodec *codec)
 {
-  FsCodec *lookup_codec = codec_copy_without_config (codec);
-  CodecAssociation *ca = NULL;
+  GList *item;
 
-  while (codec_associations)
+  for (item = codec_associations; item; item = g_list_next (item))
   {
-    CodecAssociation *tmpca = codec_associations->data;
-    FsCodec *tmpcodec;
+    CodecAssociation *ca = item->data;
 
-    if (!codec_association_is_valid_for_sending (tmpca, FALSE))
-      continue;
-
-    tmpcodec = codec_copy_without_config (tmpca->codec);
-
-    if (fs_codec_are_equal (tmpcodec, lookup_codec))
-      ca = tmpca;
-    fs_codec_destroy (tmpcodec);
-    if (ca)
-      break;
-
-    codec_associations = g_list_next (codec_associations);
+    if (codec_association_is_valid_for_sending (ca, FALSE) &&
+        fs_codec_are_equal (ca->codec, codec))
+      return ca;
   }
 
-  fs_codec_destroy (lookup_codec);
-
-  return ca;
+  return NULL;
 }
