@@ -914,9 +914,7 @@ fs_nice_transmitter_set_sending (FsNiceTransmitter *self,
             GST_ERROR ("Could not link nicesink to its tee pad");
           gst_object_unref (elempad);
 
-          gst_element_send_event (ns->nicesinks[c],
-              gst_event_new_custom (GST_EVENT_CUSTOM_UPSTREAM,
-                  gst_structure_new ("GstForceKeyUnit", NULL)));
+          fs_nice_transmitter_request_keyunit (self, ns, c);
         }
       }
     }
@@ -932,3 +930,15 @@ fs_nice_transmitter_set_sending (FsNiceTransmitter *self,
   g_mutex_unlock (ns->mutex);
 
 }
+
+
+void
+fs_nice_transmitter_request_keyunit (FsNiceTransmitter *self,
+    NiceGstStream *ns, guint component)
+{
+  g_assert (component <= self->components);
+
+  gst_element_send_event (ns->nicesinks[component],
+      gst_event_new_custom (GST_EVENT_CUSTOM_UPSTREAM,
+          gst_structure_new ("GstForceKeyUnit", NULL)));
+}
